from typing import Any

from fastapi import FastAPI
from opentelemetry import metrics, trace
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
    OTLPMetricExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter,
)
from opentelemetry.instrumentation.fastapi import (  # type: ignore
    FastAPIInstrumentor,
)
from opentelemetry.instrumentation.httpx import (  # type: ignore
    HTTPXClientInstrumentor,
)
from opentelemetry.instrumentation.logging import (  # type: ignore
    LoggingInstrumentor,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import (
    PeriodicExportingMetricReader,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from typing_extensions import Dict

from .settings import get_settings

tracer: trace.Tracer | None = None
meter: metrics.Meter | None = None


def get_tracer() -> trace.Tracer:
    if tracer is None:
        raise Exception("Tracer is not initialized")
    return tracer


def get_meter() -> metrics.Meter:
    if meter is None:
        raise Exception("Meter is not initialized")
    return meter


def get_resource_attributes() -> Dict[str, Any]:
    resources = Resource.create()
    resources_dict: Dict[str, Any] = {}
    for key in resources.attributes:
        resources_dict[key] = resources.attributes[key]
    settings = get_settings()
    if settings is None:
        raise Exception("Settings are not initialized")
    resources_dict["workspace"] = settings.workspace
    resources_dict["service.name"] = settings.name
    return resources_dict


def get_metrics_exporter() -> OTLPMetricExporter:
    return OTLPMetricExporter()


def get_span_exporter() -> OTLPSpanExporter:
    return OTLPSpanExporter()


def instrument_app(app: FastAPI):
    global tracer
    global meter
    settings = get_settings()
    if settings is None:
        raise Exception("Settings are not initialized")
    resource = Resource.create(
        {
            "service.name": settings.name,
            "service.namespace": settings.workspace,
            "service.workspace": settings.workspace,
        }
    )
    # Set up the TracerProvider
    trace_provider = TracerProvider(resource=resource)
    span_processor = BatchSpanProcessor(get_span_exporter())
    trace_provider.add_span_processor(span_processor)
    trace.set_tracer_provider(trace_provider)
    tracer = trace_provider.get_tracer(__name__)

    metrics_exporter = PeriodicExportingMetricReader(get_metrics_exporter())
    meter_provider = MeterProvider(
        resource=resource, metric_readers=[metrics_exporter]
    )
    metrics.set_meter_provider(meter_provider)
    meter = meter_provider.get_meter(__name__)

    FastAPIInstrumentor.instrument_app(  # type: ignore
        app=app, tracer_provider=trace_provider, meter_provider=meter_provider
    )
    HTTPXClientInstrumentor().instrument(meter_provider=meter_provider)  # type: ignore
    LoggingInstrumentor(tracer_provider=trace_provider).instrument(
        set_logging_format=True
    )



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="AgentRelease")


@_attrs_define
class AgentRelease:
    """Agent release, used to deploy a agent from one environment to another

    Attributes:
        from_ (Union[Unset, str]): Origin environment from which the agent is released
        to (Union[Unset, str]): Destination environment to which the agent is released
    """

    from_: Union[Unset, str] = UNSET
    to: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from_ = self.from_

        to = self.to

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if from_ is not UNSET:
            field_dict["from"] = from_
        if to is not UNSET:
            field_dict["to"] = to

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        from_ = d.pop("from", UNSET)

        to = d.pop("to", UNSET)

        agent_release = cls(
            from_=from_,
            to=to,
        )

        agent_release.additional_properties = d
        return agent_release

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="StoreFunctionParameter")


@_attrs_define
class StoreFunctionParameter:
    """Store function parameter

    Attributes:
        description (Union[Unset, str]): Store function parameter description
        name (Union[Unset, str]): Store function parameter name
        required (Union[Unset, bool]): Store function parameter required
        type_ (Union[Unset, str]): Store function parameter type
    """

    description: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    required: Union[Unset, bool] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        description = self.description

        name = self.name

        required = self.required

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if description is not UNSET:
            field_dict["description"] = description
        if name is not UNSET:
            field_dict["name"] = name
        if required is not UNSET:
            field_dict["required"] = required
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        description = d.pop("description", UNSET)

        name = d.pop("name", UNSET)

        required = d.pop("required", UNSET)

        type_ = d.pop("type", UNSET)

        store_function_parameter = cls(
            description=description,
            name=name,
            required=required,
            type_=type_,
        )

        store_function_parameter.additional_properties = d
        return store_function_parameter

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio
import urllib.parse
import warnings
from typing import Any, Callable

import pydantic
import pydantic_core
import typing_extensions as t
from beamlit.api.functions import get_function
from beamlit.authentication.authentication import AuthenticatedClient
from beamlit.models.function import Function
from langchain_core.tools.base import BaseTool, BaseToolkit, ToolException
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema as cs


def create_schema_model(schema: dict[str, t.Any]) -> type[pydantic.BaseModel]:
    # Create a new model class that returns our JSON schema.
    # LangChain requires a BaseModel class.
    class Schema(pydantic.BaseModel):
        model_config = pydantic.ConfigDict(extra="allow")

        @t.override
        @classmethod
        def __get_pydantic_json_schema__(
            cls, core_schema: cs.CoreSchema, handler: pydantic.GetJsonSchemaHandler
        ) -> JsonSchemaValue:
            return schema

    return Schema


class RemoteTool(BaseTool):
    """
    Remote tool
    """

    client: AuthenticatedClient
    handle_tool_error: bool | str | Callable[[ToolException], str] | None = True

    @t.override
    def _run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        warnings.warn(
            "Invoke this tool asynchronousely using `ainvoke`. This method exists only to satisfy standard tests.",
            stacklevel=1,
        )
        return asyncio.run(self._arun(*args, **kwargs))

    @t.override
    async def _arun(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        result = self.client.call_tool(self.name, arguments=kwargs)
        response = result.json()
        content = pydantic_core.to_json(response["content"]).decode()
        if response["isError"]:
            raise ToolException(content)
        return content

    @t.override
    @property
    def tool_call_schema(self) -> type[pydantic.BaseModel]:
        assert self.args_schema is not None  # noqa: S101
        return self.args_schema

class RemoteToolkit(BaseToolkit):
    """
    Remote toolkit
    """

    client: AuthenticatedClient
    function: str
    _function: Function | None = None

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    def initialize(self) -> None:
        """Initialize the session and retrieve tools list"""
        if self._function is None:
            self._function = get_function(self.function, client=self.client)

    @t.override
    def get_tools(self) -> list[BaseTool]:
        if self._tools is None:
            raise RuntimeError("Must initialize the toolkit first")

        if self._function.spec.kit:
            return [
                RemoteTool(
                client=self.client,
                name=func.name,
                description=func.description or "",
                args_schema=create_schema_model(func.parameters),
                )
                for func in self._function.spec.kit
            ]

        return [
            RemoteTool(
                client=self.client,
                name=self._function.metadata.name,
                description=self._function.spec.description or "",
                args_schema=create_schema_model(self._function.spec.parameters),
            )
        ]


from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CreateWorkspaceServiceAccountResponse200")


@_attrs_define
class CreateWorkspaceServiceAccountResponse200:
    """
    Attributes:
        client_id (Union[Unset, str]): Service account client ID
        client_secret (Union[Unset, str]): Service account client secret (only returned on creation)
        created_at (Union[Unset, str]): Creation timestamp
        description (Union[Unset, str]): Service account description
        name (Union[Unset, str]): Service account name
        updated_at (Union[Unset, str]): Last update timestamp
    """

    client_id: Union[Unset, str] = UNSET
    client_secret: Union[Unset, str] = UNSET
    created_at: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    updated_at: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        client_id = self.client_id

        client_secret = self.client_secret

        created_at = self.created_at

        description = self.description

        name = self.name

        updated_at = self.updated_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if client_id is not UNSET:
            field_dict["client_id"] = client_id
        if client_secret is not UNSET:
            field_dict["client_secret"] = client_secret
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if description is not UNSET:
            field_dict["description"] = description
        if name is not UNSET:
            field_dict["name"] = name
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        client_id = d.pop("client_id", UNSET)

        client_secret = d.pop("client_secret", UNSET)

        created_at = d.pop("created_at", UNSET)

        description = d.pop("description", UNSET)

        name = d.pop("name", UNSET)

        updated_at = d.pop("updated_at", UNSET)

        create_workspace_service_account_response_200 = cls(
            client_id=client_id,
            client_secret=client_secret,
            created_at=created_at,
            description=description,
            name=name,
            updated_at=updated_at,
        )

        create_workspace_service_account_response_200.additional_properties = d
        return create_workspace_service_account_response_200

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Generator

from httpx import Auth, Request, Response


class ApiKeyProvider(Auth):
    def __init__(self, credentials, workspace_name: str):
        self.credentials = credentials
        self.workspace_name = workspace_name

    def get_headers(self):
        return {
            "X-Beamlit-Api-Key": self.credentials.apiKey,
            "X-Beamlit-Workspace": self.workspace_name,
        }

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers["X-Beamlit-Api-Key"] = self.credentials.apiKey
        request.headers["X-Beamlit-Workspace"] = self.workspace_name
        yield request
